const item_names = ["UNITS", "TIME", "TIMESTEP", "NUMBER OF ATOMS", "BOX BOUNDS", "ATOMS"]
const optional_item_names = ["UNITS", "TIME"]

function readnextframe(s::MDFileStream, ::Val{LAMMPSDump}, ::Type{T})::MDFrame{T} where T
    check_eof(s)
    
    timestep::Int = 0
    natoms::Int = 0
    time::Union{T,Nothing} = nothing
    units::Union{String,Nothing} = nothing
    box::Union{SimulationBox{T},Nothing} = nothing
    particles::Union{DataFrame,Nothing} = nothing
    
    itemheader, itemname = readitemheader(s)
    for (i, name_i) in enumerate(item_names)
        if itemname != name_i
            if !(name_i in optional_item_names)
                throw(FileParsingError(s, linenumber(s)-1, "Expected item $name_i"))
            end
            continue
        end
        
        if itemname == "UNITS"
            units = readline(s)
        elseif itemname == "TIME"
            time = parseline(s, T)
        elseif itemname == "TIMESTEP"
            timestep = parseline(s, Int)
        elseif itemname == "NUMBER OF ATOMS"
            natoms = parseline(s, Int)
        elseif itemname == "BOX BOUNDS"
            box = readbox(s, T, itemheader)
        elseif itemname == "ATOMS"
            particles = readparticles(s, T, itemheader, natoms)
        end
        
        if i != length(item_names)
            itemheader, itemname = readitemheader(s)
        end
    end
    
    MDFrame{T}(timestep, box, particles)
end

function readitemheader(s::MDFileStream)::Tuple{String,String}
    check_eof(s)
    header = readline(s)
    try
        header, itemname(header)
    catch ex
        throw(FileParsingError(s, linenumber(s)-1, ex.msg))
    end
end

function itemname(itemheader::String)::String
    for name in item_names
        pref = "ITEM: " * name
        if length(itemheader) > length(pref)
            pref *= " "
        end
        if startswith(itemheader, pref)
            return name
        end
    end
    throw(ArgumentError("Unrecognized item"))
end

function parseline(s::MDFileStream, ::Type{T}) where T
    check_eof(s)
    try
        return parse(T, readline(s))
    catch ex
        throw(FileParsingError(s, linenumber(s)-1, ex.msg))
    end
end

function readbox(s::MDFileStream, ::Type{T}, itemheader::String)::SimulationBox{T} where T<:AbstractFloat
    pref = "ITEM: BOX BOUNDS "
    itemheader = strip(itemheader[length(pref)+1:length(itemheader)])    
    
    bcstr = ""
    triclinic = false
    if startswith(itemheader, "xy xz yz")
        triclinic = true
        bcstr = itemheader[length("xy xz yz")+2:length(itemheader)]
    else
        bcstr = itemheader
    end
    
    bc = try
        parse(BoundaryConditions, bcstr)
    catch ex
        throw(FileParsingError(s, linenumber(s)-1, ex.msg))
    end
    
    boxdf::Union{DataFrame,Nothing} = nothing
    
    if triclinic
        boxdf = readtable(s; rows=3, header=["lo", "hi", "tilt"], types=[T, T, T])
    else
        boxdf = readtable(s; rows=3, header=["lo", "hi"], types=[T, T])
    end
    
    basis = zeros(MMatrix{3,3,T})
    origin = zeros(MVector{3,T})
    
    for i in 1:3
        basis[i,i] = boxdf.hi[i] - boxdf.lo[i]
        origin[i] = boxdf.lo[i]
    end
    
    if triclinic
        basis[1,2] = boxdf.tilt[1]
        basis[1,3] = boxdf.tilt[2]
        basis[2,3] = boxdf.tilt[3]
    end
    
    try
        SimulationBox{T}(basis, origin, bc)
    catch ex
        throw(FileParsingError(s, linenumber(s)-3, ex.msg))
    end
end

attribute_types = Dict(
    "id" => Int,
    "mol" => Int,
    "proc" => Int,
    "type" => Int,
    "element" => String
)

function readparticles(s::MDFileStream, ::Type{T}, itemheader::String, natoms::Int)::DataFrame where T<:AbstractFloat
    pref = "ITEM: ATOMS "
    itemheader = strip(itemheader[length(pref)+1:length(itemheader)])
    tableheader = map(string, split(itemheader))
    types = DataType[]
    
    for attribute in tableheader
        if attribute in keys(attribute_types)
            push!(types, attribute_types[attribute])
        else
            push!(types, T)
        end
    end
    
    readtable(s; rows=natoms, header=tableheader, types=types)
end

function readtable(s::MDFileStream; rows::Int, kwargs...)
    line = linenumber(s)
    v, n = readlines(s, rows)
    if n != rows
        throw(unexpected_eof(s))
    end
    try
        readtable(v; kwargs...)
    catch ex
        if ex isa TableParsingError
            line += ex.row - 1
        end
        throw(FileParsingError(s, line, sprint(showerror, ex)))
    end
end

function Base.parse(::Type{BoundaryConditions}, str::AbstractString)
    s = split(str)
    if length(s) != 3
        throw(ArgumentError("Expected boundary conditions for each axis: $str"))
    end
    xlo, xhi = parse(NTuple{2,BoundaryCondition}, s[1])
    ylo, yhi = parse(NTuple{2,BoundaryCondition}, s[2])
    zlo, zhi = parse(NTuple{2,BoundaryCondition}, s[3])
    BoundaryConditions(xlo, xhi, ylo, yhi, zlo, zhi)
end

function Base.parse(::Type{NTuple{2,BoundaryCondition}}, str::AbstractString)
    if length(str) == 1
        return (parse(BoundaryCondition, str), parse(BoundaryCondition, str))
    elseif length(str) == 2
        return (parse(BoundaryCondition, str[1:1]), parse(BoundaryCondition, str[2:2]))
    else
        throw(ArgumentError("Boundary conditions axis string must be 1 or 2 characters long: $str"))
    end
end

function Base.parse(::Type{BoundaryCondition}, str::AbstractString)
    if str == "p"
        return periodic
    elseif str == "f"
        return fixed
    elseif str == "s"
        return shrinkwrap
    elseif str == "m"
        return minshrinkwrap
    else
        throw(ArgumentError("Invalid boundary condition character $str (expected one of p, f, s, m)"))
    end
end

function check_eof(s::MDFileStream)
    if eof(s)
        throw(unexpected_eof(s))
    end
end

