struct MDFrame{T<:AbstractFloat}
    timestep::Int
    box::SimulationBox{T}
    particles::DataFrame
    properties::Dict{Symbol,Any}
end

function MDFrame{T}(timestep::Int, box::SimulationBox{T}, particles::DataFrame) where T <: AbstractFloat
    sort!(particles, [:id])
    MDFrame{T}(timestep, box, particles, Dict{Symbol,Any}())
end

function timestep(frame::MDFrame{T})::Int where T<:AbstractFloat
    frame.timestep
end

function Base.length(frame::MDFrame{T})::Int where T<:AbstractFloat
    nrow(frame.particles)
end

function positions(frame::MDFrame{T})::Vector{SVector{3,T}} where T<:AbstractFloat
    if has_positions(frame)
        _hcat3(frame.particles.x, frame.particles.y, frame.particles.z, T)
    elseif has_scaled_positions(frame)
        scaled_positions = _hcat3(frame.particles.xs, frame.particles.ys, frame.particles.zs, T)
        scaled_to_unscaled(box(frame), scaled_positions)
    elseif has_scaled_unwrapped_positions(frame)
        scaled_unwrapped_positions = _hcat3(frame.particles.xsu, frame.particles.ysu, frame.particles.zsu, T)
        scaled_positions = wrap_scaled_positions(box(frame), scaled_unwrapped_positions)
        scaled_to_unscaled(box(frame), scaled_positions)
    elseif has_unwrapped_positions(frame)
        unwrapped_positions = _hcat3(frame.particles.xu, frame.particles.yu, frame.particles.zu, T)
        scaled_unwrapped_positions = unscaled_to_scaled(box(frame), unwrapped_positions)
        scaled_positions = wrap_scaled_positions(box(frame), scaled_unwrapped_positions)
        scaled_to_unscaled(box(frame), scaled_positions)
    else
        throw(ArgumentError("Could not compute particle wrapped positions from the frame data"))
    end
end

function spositions(frame::MDFrame{T})::Vector{SVector{3,T}} where T<:AbstractFloat
    cols = names(frame.particles)
    if "xs" in cols && "ys" in cols && "zs" in cols
        return _hcat3(frame.particles.xs, frame.particles.ys, frame.particles.zs, T)
    end
    
    try
        su = supositions(frame)
        wrap_scaled_positions(box(frame), su)
    catch ex
        throw(ArgumentError("Could not compute particle scaled positions from the frame data"))
    end
end

function supositions(frame::MDFrame{T})::Vector{SVector{3,T}} where T<:AbstractFloat
    cols = names(frame.particles)
    if "xsu" in cols && "ysu" in cols && "zsu" in cols
        return _hcat3(frame.particles.xsu, frame.particles.ysu, frame.particles.zsu, T)
    end
    try
        u = upositions(frame)
        unscaled_to_scaled(box(frame), u)
    catch ex
        throw(ArgumentError("Could not compute particle scaled unwrapped positions from the frame data"))
    end
end

function upositions(frame::MDFrame{T})::Vector{SVector{3,T}} where T<:AbstractFloat
    cols = names(frame.particles)
    if "xu" in cols && "yu" in cols && "zu" in cols
        return _hcat3(frame.particles.xu, frame.particles.yu, frame.particles.zu, T)
    end
    throw(ArgumentError("No unwrapped position data in the frame"))
end

function velocities(frame::MDFrame{T})::Vector{SVector{3,T}} where T<:AbstractFloat
    cols = names(frame.particles)
    if "vx" in cols && "vy" in cols && "vz" in cols
        return _hcat3(frame.particles.vx, frame.particles.vy, frame.particles.vz, T)
    end
    throw(ArgumentError("No velocity data in the frame"))
end

function images(frame::MDFrame{T})::Vector{SVector{3,Int16}} where T<:AbstractFloat
    cols = names(frame.particles)
    if has_images(frame)
        return _hcat3(frame.particles.ix, frame.particles.iy, frame.particles.iz, T)
    else
        scaled_unwrapped_positions = try
            scaled_unwrapped_positions(frame)
        catch
            throw(ArgumentError("No image data in the frame and no unwrapped positions, can't compute image data"))
        end
        [Int16.(floor.(v)) for v in scaled_unwrapped_positions]
    end
end

function id_tags(frame::MDFrame{T})::Vector{Int} where T<:AbstractFloat
    frame.particles.id
end

function particle_types(frame::MDFrame{T})::Vector{Int} where T<:AbstractFloat
    frame.particles.type
end

function mol_tags(frame::MDFrame{T})::Vector{Int} where T<:AbstractFloat
    frame.particles.mol
end

function extra_scalars(frame::MDFrame{T})::Dict{Symbol, Vector{T}} where T<:AbstractFloat
    extra_scalars = Dict{Symbol, Vector{T}}()
    exclude_set = Set([:id, :mol, :type, :x, :y, :z, :xs, :ys, :zs, :xu, :yu, :zu, :xsu, :ysu, :zsu, :ix, :iy, :iz, :vx, :vy, :vz])
    for name in propertynames(frame.particles)
        if !(name in exclude_set)
            extra_scalars[name] = frame.particles[!, name]
        end
    end
    extra_scalars
end

function _hcat3(v1::AbstractVector, v2::AbstractVector, v3::AbstractVector, ::Type{T})::Vector{SVector{3,T}} where T
    vt1::Vector{T} = v1
    vt2::Vector{T} = v2
    vt3::Vector{T} = v3
    n = length(v1)
    [SVector{3,T}(vt1[i], vt2[i], vt3[i]) for i in 1:n]
end

function boxvectors(frame::MDFrame{T})::SMatrix{3,3,T} where T<:AbstractFloat
    basis(box(frame))
end

function boxorigin(frame::MDFrame{T})::SVector{3,T} where T<:AbstractFloat
    origin(box(frame))
end

function pbc(frame::MDFrame{T})::NTuple{3,Bool} where T<:AbstractFloat
    pbc(box(frame))
end

function box(frame::MDFrame{T})::SimulationBox{T} where T<:AbstractFloat
    frame.box
end

function has_positions(frame::MDFrame{T}) where T<:AbstractFloat
    cols = names(frame.particles)
    "x" in cols && "y" in cols && "z" in cols
end

function has_scaled_positions(frame::MDFrame{T}) where T<:AbstractFloat
    cols = names(frame.particles)
    "xs" in cols && "ys" in cols && "zs" in cols
end

function has_unwrapped_positions(frame::MDFrame{T}) where T<:AbstractFloat
    cols = names(frame.particles)
    "xu" in cols && "yu" in cols && "zu" in cols
end

function has_scaled_unwrapped_positions(frame::MDFrame{T}) where T<:AbstractFloat
    cols = names(frame.particles)
    "xsu" in cols && "ysu" in cols && "zsu" in cols
end

function has_images(frame::MDFrame{T}) where T<:AbstractFloat
    cols = names(frame.particles)
    "ix" in cols && "iy" in cols && "iz" in cols
end