@enum BoundaryCondition begin
    periodic
    fixed
    shrinkwrapped
    minshrinkwrapped
end

struct BoundaryConditions
    xlo::BoundaryCondition
    xhi::BoundaryCondition
    ylo::BoundaryCondition
    yhi::BoundaryCondition
    zlo::BoundaryCondition
    zhi::BoundaryCondition
    
    function BoundaryConditions(xlo, xhi, ylo, yhi, zlo, zhi)
        checkperiodicity(xlo, xhi, "x")
        checkperiodicity(ylo, yhi, "y")
        checkperiodicity(zlo, zhi, "z")
        new(xlo, xhi, ylo, yhi, zlo, zhi)
    end
end

function checkperiodicity(lo::BoundaryCondition, hi::BoundaryCondition, axis::String)
    if (lo === periodic) ⊻ (hi === periodic)
        throw(ArgumentError("Both or none of the $axis axis boundaries may be periodic: $lo $hi"))
    end
end

function pbc(bc::BoundaryConditions)::NTuple{3,Bool}
    bc.xlo === periodic, bc.ylo === periodic, bc.zlo === periodic
end

struct SimulationBox{T<:AbstractFloat}
    basis::SMatrix{3,3,T}
    origin::SVector{3,T}
    bc::BoundaryConditions
    
    function SimulationBox{T}(basis, origin, bc) where T 
        if det(basis) == 0
            throw(ArgumentError("$basis is not a valid basis (matrix is singular)"))
        end
        new{T}(basis, origin, bc)
    end
end

function pbc(box::SimulationBox)::NTuple{3,Bool}
    pbc(box.bc)
end

function basis(box::SimulationBox{T})::SMatrix{3,3,T} where T
    box.basis
end

function origin(box::SimulationBox{T})::SVector{3,T} where T
    box.origin
end

function avec(box::SimulationBox{T})::SVector{3,T} where T
    box.basis[:,1]
end

function bvec(box::SimulationBox{T})::SVector{3,T} where T
    box.basis[:,2]
end

function cvec(box::SimulationBox{T})::SVector{3,T} where T
    box.basis[:,3]
end

function scaled_to_unscaled(box::SimulationBox{T}, scaled_positions::Vector{SVector{3,T}})::Vector{SVector{3,T}} where T<:AbstractFloat
    b = basis(box)
    o = origin(box)
    [o + b * v for v in scaled_positions]
end

function unscaled_to_scaled(box::SimulationBox{T}, unscaled_positions::Vector{SVector{3,T}})::Vector{SVector{3,T}} where T<:AbstractFloat
    bi = basis(box) ^ -1
    o = origin(box)
    [bi * (v - o) for v in unscaled_positions]
end

function wrap_scaled_positions(box::SimulationBox{T}, scaled_unwrapped_positions::Vector{SVector{3,T}})::Vector{SVector{3,T}} where T<:AbstractFloat
    [SVector{3,T}(v[1] - floor(v[1]), v[2] - floor(v[2]), v[3] - floor(v[3])) for v in scaled_unwrapped_positions]
end
